{-
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
-}
module DpTransImpl where {

import Util;
import List;
import Trs;
import Term;
import HOL;
import ErrorMonad;
import Show;

varsLI :: forall a b. (Eq b) => Term a b -> [b];
varsLI (Fun f ss) = foldra (\ s -> unionI (varsLI s)) ss [];
varsLI (Var x) = [x];

checkSupterm :: forall a b. (Eq a, Eq b) => Term a b -> Term a b -> Bool;
checkSupterm (Fun f ts) s = member s ts || testPos (\ t -> checkSupterm t s) ts;
checkSupterm (Var x) uu = False;

computeDPsOfTerm ::
  forall a b.
    (Eq a,
      Eq b) => [a] ->
                 Term a b ->
                   Term (Shp a) b ->
                     Term a b -> [(Term (Shp a) b, Term (Shp a) b)];
computeDPsOfTerm d l lshp (Fun f ts) =
  append
    (if member f d && not (checkSupterm l (Fun f ts))
      then [(lshp, sharpa (Fun f ts))] else [])
    (foldra (\ t -> unionI (computeDPsOfTerm d l lshp t)) ts []);
computeDPsOfTerm d l lshp (Var uu) = [];

computeDPsOfRule ::
  forall a b.
    (Eq a,
      Eq b) => [a] ->
                 (Term a b, Term a b) -> [(Term (Shp a) b, Term (Shp a) b)];
computeDPsOfRule d (Var x, r) = [];
computeDPsOfRule d (Fun f ss, r) =
  computeDPsOfTerm d (Fun f ss) (sharpa (Fun f ss)) r;

computeDSymbs :: forall a b. (Eq a) => [(Term a b, Term a b)] -> [a];
computeDSymbs ((Var uw, ux) : ts) = computeDSymbs ts;
computeDSymbs ((Fun f uu, uv) : ts) =
  let {
    fs = computeDSymbs ts;
  } in (if member f fs then fs else f : fs);
computeDSymbs [] = [];

computeDPs ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] -> [(Term (Shp a) b, Term (Shp a) b)];
computeDPs r =
  let {
    d = computeDSymbs r;
  } in foldra (\ lr -> unionI (computeDPsOfRule d lr)) r [];

checkWfTRS ::
  forall f v.
    (Sshow f, Eq v, Sshow v) => [(Term f v, Term f v)] -> Eithera [[Char]] ();
checkWfTRS r =
  catcha
    (bind (catcha
            (checkAllMI
              (\ lr ->
                checkB (not (isVar (fst lr)))
                  (concata
                    ["the left side of rule ", showRule lr, " is a variable"]))
              r)
            (\ x -> errora (snd x)))
      (\ x ->
        catcha
          (checkAllMI
            (\ lr ->
              catcha (checkSubset (varsLI (snd lr)) (varsLI (fst lr)))
                (\ xa ->
                  errora
                    (concata
                      ["the variable ", sshow xa,
                        " occurs only in the right-hand side of rule ",
                        showRule lr])))
            r)
          (\ xa -> errora (snd xa))))
    (\ x -> errora ["the TRS is not well-formed", x]);

checkDPs ::
  forall f v.
    (Eq f, Sshow f, Eq v,
      Sshow v) => [(Term f v, Term f v)] ->
                    [(Term (Shp f) v, Term (Shp f) v)] -> Eithera [[Char]] ();
checkDPs r p =
  catcha
    (bind (checkWfTRS r)
      (\ x ->
        catcha (checkSubset (computeDPs r) p)
          (\ xa ->
            errora
              [concata
                 ["the DP ", showRule xa,
                   " does not appear in the DP problem"]])))
    (\ x -> errora ("the DP-transformation is not applied correctly." : x));

computeID ::
  forall a b. [(Term a b, Term a b)] -> [(Term (Shp a) b, Term (Shp a) b)];
computeID r = mapa (\ lr -> (plaina (fst lr), plaina (snd lr))) r;

}
