(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Polynomial imports RedPair Show
begin

types carrier = nat
  
datatype 'v poly = Poly carrier "(carrier \<times> 'v) list" (* currently only linear polys *)

instantiation poly :: (sshow) sshow
begin
fun show_pair :: "(carrier \<times> 'v :: sshow) \<Rightarrow> sstring" 
where "show_pair (0,v) = []"
   |  "show_pair (Suc 0,v) = (sshow v)"
   |  "show_pair (Suc (Suc n),v) = (sshow (Suc (Suc n)) +++ ''*'' +++ sshow v)"
fun show_number :: "carrier \<Rightarrow> sstring" 
where "show_number 0 = []"
   |  "show_number (Suc n) = sshow (Suc n)"
fun show_entries :: "('v :: sshow)poly \<Rightarrow> sstring list"
where "show_entries (Poly c vp) = (filter (\<lambda> entry. entry \<noteq> []) (show_number c # map show_pair vp))"
fun merge_entries :: "sstring list \<Rightarrow> sstring"
where "merge_entries [] = []"
    | "merge_entries (x # xs) = ('' + '' +++ x +++ (merge_entries xs))"
fun sshow_poly :: "('v :: sshow)poly \<Rightarrow> sstring"
 where "sshow_poly p = (case  show_entries p of Nil \<Rightarrow> ''0''
                             |      (Cons d es) \<Rightarrow> (d +++ (merge_entries es)))"
instance ..
end



abbreviation polyZero :: "'v poly" where "polyZero \<equiv> Poly 0 []"


fun lookupVar :: "'v \<Rightarrow> (carrier \<times> 'v) list \<Rightarrow> (carrier \<times> (carrier \<times> 'v) list) option"
  where "lookupVar x [] = None"
    |   "lookupVar x ((c,y) # vp) = (if x = y then Some (c, vp) else 
             (case lookupVar x vp of None \<Rightarrow> None | Some (d,vpp) \<Rightarrow> Some (d, (c,y) # vpp)))"

fun sumPolyVars :: "(carrier \<times> 'v) list \<Rightarrow> (carrier \<times> 'v) list \<Rightarrow> (carrier \<times> 'v) list"
  where "sumPolyVars [] vp2 = vp2"
     |  "sumPolyVars ((c,x) # vp1) vp2 = (case lookupVar x vp2 of 
                  None \<Rightarrow> (c,x) # sumPolyVars vp1 vp2 
                | Some (d, vpp2) \<Rightarrow> (c + d, x) # sumPolyVars vp1 vpp2)" 

fun polySum :: "'v poly \<Rightarrow> 'v poly \<Rightarrow> 'v poly" (infix "+" 65)
  where "(Poly c vp1) + (Poly d vp2) = Poly (c+d) (sumPolyVars vp1 vp2)"

fun polyEval :: "('v \<Rightarrow> carrier ) \<Rightarrow> 'v poly \<Rightarrow> carrier" 
  where "polyEval \<beta> (Poly c vp) = (c + listsum (map (\<lambda> nv. fst nv * (\<beta> (snd nv))) vp))"

fun polyMult :: "carrier \<Rightarrow> 'v poly \<Rightarrow> 'v poly" (infix "*" 70)
  where "n * (Poly c vp) = (Poly (n * c) (map (\<lambda> cx. (n * fst cx, snd cx)) vp))"



definition polyOne :: "'v poly" where "polyOne \<equiv> Poly 1 []"
lemma [simp] : "polyEval \<beta> polyOne = 1" unfolding polyOne_def by simp

lemma polySumSound[simp] : "polyEval \<beta> (polySum p q) = polyEval \<beta> p + polyEval \<beta> q"
proof (cases p)
  case (Poly c vp)
  show ?thesis
  proof (cases q)
    case (Poly d vq)
    have "polyEval \<beta> (polySum (Poly c vp) (Poly d vq)) = polyEval \<beta> (Poly c vp) + polyEval \<beta> (Poly d vq)"
    proof (simp)
      show "(\<Sum>nv\<leftarrow>sumPolyVars vp vq. fst nv * \<beta> (snd nv)) =
    (\<Sum>nv\<leftarrow>vp. fst nv * \<beta> (snd nv)) + (\<Sum>nv\<leftarrow>vq. fst nv * \<beta> (snd nv))"
      proof (induct vp arbitrary: vq)
	case Nil thus ?case by simp
      next
	case (Cons nv vpp)
	show ?case
	proof (cases nv)
	  case (Pair n x)
	  have "(\<Sum>nv\<leftarrow>sumPolyVars ((n,x) # vpp) vq. fst nv * \<beta> (snd nv)) =
          (\<Sum>nv\<leftarrow>(n,x) # vpp. fst nv * \<beta> (snd nv)) +
          (\<Sum>nv\<leftarrow>vq. fst nv * \<beta> (snd nv))"
	  proof (cases "lookupVar x vq")
	    case None
	    with Cons show ?thesis by simp
	  next
	    case (Some evq)
	    show ?thesis 
	    proof (cases evq)
	      case (Pair e vqq)
	      with Some have lookup: "lookupVar x vq = Some (e, vqq)" by simp
	      with Cons show ?thesis proof (auto simp: plus_times.distrib)
		from lookup show "e * \<beta> x + (\<Sum>nv\<leftarrow>vqq. fst nv * \<beta> (snd nv)) =
       (\<Sum>nv\<leftarrow>vq. fst nv * \<beta> (snd nv))"
		proof (induct vq arbitrary: e vqq, simp)
		  case (Cons cy vp)
		  show ?case proof (cases cy)
		    case (Pair c y)
		    with Cons  have lookup2: "lookupVar x ((c,y) # vp) = Some (e, vqq)" by simp
		    hence "e * \<beta> x + (\<Sum>nv\<leftarrow>vqq. fst nv * \<beta> (snd nv)) = (\<Sum>nv\<leftarrow>(c,y) # vp. fst nv * \<beta> (snd nv))"
		    proof (cases "x=y", simp, cases "lookupVar x vp", simp)
		      fix dvpp
		      assume xy: "x \<noteq> y" and lookup3: "lookupVar x vp = Some dvpp"
		      show ?thesis
		      proof (cases dvpp)
			case (Pair d vpp)
			with lookup3 have lookup4: "lookupVar x vp = Some (d, vpp)" by simp
			with lookup2 Cons xy show ?thesis by auto
		      qed
		    qed
		    with Pair show ?thesis by simp
		  qed
		qed
	      qed
	    qed
	  qed			    
	  with Pair show ?thesis by simp
	qed
      qed
    qed
    with prems show ?thesis by simp
  qed
qed




fun polyC :: "'v poly \<Rightarrow> carrier" where "polyC (Poly c _) = c"
fun polyVP :: "'v poly \<Rightarrow> (carrier \<times> 'v) list" where "polyVP (Poly _ vp) = vp"


types 'f "polyinter" = "'f \<Rightarrow> carrier \<times> (carrier list)"

(*
locale polyorder 
begin
*)
function polyOfTerm :: "'f polyinter \<Rightarrow> ('f,'v) term \<Rightarrow> 'v poly" ("\<guillemotleft>_,_\<guillemotright>" 50)
where "\<guillemotleft>\<alpha>,Var x\<guillemotright> = (Poly 0 [(1,x)])"
   |  "\<guillemotleft>\<alpha>,Fun f ts\<guillemotright> = (let (c, list) = \<alpha> f in polySum (Poly c []) (foldr polySum
               (map (\<lambda> ct. (fst ct) *  (\<guillemotleft>\<alpha>,snd ct\<guillemotright>)) (zip list ts))                 
                        (Poly 0 [])))"
by (pat_completeness, auto)

termination polyOfTerm
proof
  fix \<alpha> f and ts :: "('f,'v) term list" and cl const list ct
  assume "ct \<in> set (zip list ts)"
  hence "ct mem (zip list ts)" by (auto simp: mem_iff)
  hence "snd ct mem ts" proof (induct ts arbitrary: list, simp)
    case (Cons a ts)
    show ?case 
    proof (cases list)
      case Nil with Cons show ?thesis by simp
    next
      case (Cons b ll)
      from prems show ?thesis by (cases "ct = (b,a)", auto)
    qed
  qed
  thus "((\<alpha>,snd ct), (\<alpha>, Fun f ts)) \<in> measures [size \<circ> snd]" by (auto simp: mem_iff)
next
  show "wf (measures [size \<circ> snd])" by simp
qed




lemma polyMultSound[simp] : "polyEval \<beta> (c * p) = c * polyEval \<beta> p"
proof (cases p)
  case (Poly d vp)
  have "polyEval \<beta> (c * Poly d vp) = c * polyEval \<beta> (Poly d vp)" 
  proof simp
    let ?left = "\<Sum>nv\<leftarrow>map (\<lambda> cx. (c * fst cx , snd cx)) vp. fst nv * \<beta> (snd nv)"
    let ?right = "\<Sum>nv\<leftarrow>vp. fst nv * \<beta> (snd nv)"
    have "?left = c * ?right" 
    proof (induct vp)
       case Nil thus ?case by simp
    next
      case (Cons nv vpp)
      thus ?case 
      proof (cases nv)
	case (Pair n v)
	with Cons show ?thesis by (auto simp: plus_times.right_distrib) 
      qed
    qed
    thus "c * d + ?left = c * (d + ?right)" by (simp add: plus_times.right_distrib)
  qed
  with Poly show ?thesis by simp
qed


definition polyGe :: "'v poly \<Rightarrow> 'v poly \<Rightarrow> bool" (infix "\<ge>" 10)
  where "(p1 \<ge> p2) = (\<forall> \<beta>. polyEval \<beta> p1 \<ge> polyEval \<beta> p2)"
definition polyGt :: "'v poly \<Rightarrow> 'v poly \<Rightarrow> bool" (infix ">" 10)
  where "(p1 > p2) = (\<forall> \<beta>. polyEval \<beta> p1 > polyEval \<beta> p2)"

definition poloS :: "'f polyinter \<Rightarrow> ('f,'v)trel"
  where "poloS \<alpha> \<equiv> {(s,t). \<guillemotleft>\<alpha>,s\<guillemotright> > \<guillemotleft>\<alpha>,t\<guillemotright>}"

definition poloNS :: "'f polyinter \<Rightarrow> ('f,'v)trel"
  where "poloNS \<alpha> \<equiv> {(s,t). \<guillemotleft>\<alpha>,s\<guillemotright>  \<ge> \<guillemotleft>\<alpha>,t\<guillemotright>}"

lemma poloSN[simp] : 
  shows "SN (poloS \<alpha> :: ('f,'v)trel)" (is "SN ?P")
proof
  fix t :: "('f,'v) term"
  show "SN ?P t"
  proof (rule ccontr)
    assume "\<not> SN ?P t"
    hence "\<exists>f . f 0 = t \<and> (\<forall>i. (f i,f(Suc i)) \<in> ?P)" by auto
    from this obtain f where "f 0 = t" and steps: "(\<forall> i. (f i, f (Suc i)) \<in> ?P)" by auto
    let ?g = "\<lambda> i. polyEval ((\<lambda> x. 0) :: 'v \<Rightarrow> carrier) (\<guillemotleft>\<alpha>, f i\<guillemotright>)"
    have "\<forall> i. ?g i > ?g (Suc i)"
    proof 
      fix i
      from steps have "(f i, f (Suc i)) \<in> poloS \<alpha>" .. 
      thus "?g i > ?g (Suc i)"  unfolding poloS_def by (auto simp: polyGt_def poloS_def poloNS_def )
    qed
    from this obtain g where nSN: "\<forall> i. ((g i) :: carrier) > g (Suc i)" by force
    have "{(x,y). (x :: carrier) < y} = ({(x,y). (x :: carrier) > y}^-1)" by auto
    with wf_less have "wf ({(x,y). (x :: carrier) > y}^-1)" by simp 
    hence "SN {(x,y). (x :: carrier) > y}" by (rule wf_imp_SN)
    with nSN show False by (auto simp: SN_ars_def SN_elt_def)
  qed
qed
  

lemma poloCompat[simp] : "poloS \<alpha> O poloNS \<alpha> \<subseteq> poloS \<alpha>"
proof
  fix st
  assume "st \<in> poloS \<alpha> O poloNS \<alpha>"
  thus "st \<in> poloS \<alpha>"
  proof (cases st, auto simp: polyGe_def poloNS_def poloS_def polyGt_def)
    fix t u s \<beta>
    assume "\<forall> \<beta>. polyEval \<beta> (\<guillemotleft>\<alpha>, s\<guillemotright>) < polyEval \<beta> (\<guillemotleft>\<alpha>, u\<guillemotright>)" and 
           "\<forall> \<beta>. polyEval \<beta> (\<guillemotleft>\<alpha>, u\<guillemotright>) \<le> polyEval \<beta> (\<guillemotleft>\<alpha>, t\<guillemotright>)"
    hence "polyEval \<beta> (\<guillemotleft>\<alpha>, s\<guillemotright>) < polyEval \<beta> (\<guillemotleft>\<alpha>, u\<guillemotright>)" and
          "polyEval \<beta> (\<guillemotleft>\<alpha>, u\<guillemotright>) \<le> polyEval \<beta> (\<guillemotleft>\<alpha>, t\<guillemotright>)" by auto
    thus "polyEval \<beta> (\<guillemotleft>\<alpha>, s\<guillemotright>) < polyEval \<beta> (\<guillemotleft>\<alpha>, t\<guillemotright>)" by simp
  qed
qed


  

lemma poloCC[simp] : 
  shows "one_step_ctxt_closed (poloNS \<alpha>  :: ('f,'v)trel)" (is "one_step_ctxt_closed ?P")
proof (unfold one_step_ctxt_closed_def, (rule allI)+, rule impI)
  fix f :: 'f 
    and bef :: "('f,'v) term list" 
    and s :: "('f,'v) term" 
    and t :: "('f,'v) term" 
    and aft :: "('f,'v) term list" 
  assume "(s, t) \<in> ?P"
  hence st: "\<guillemotleft>\<alpha>, s\<guillemotright> \<ge> \<guillemotleft>\<alpha>, t\<guillemotright>" by (simp add: poloNS_def)
  show "(Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> ?P"
  proof (cases "\<alpha> f")
    case (Pair c list)
    thus ?thesis proof (simp add: Let_def polyGe_def poloNS_def, auto)
      fix \<beta> :: "'v \<Rightarrow> carrier"
      from st have stb : "polyEval \<beta> (\<guillemotleft>\<alpha>, s\<guillemotright>) \<ge> polyEval \<beta> (\<guillemotleft>\<alpha>, t\<guillemotright>)" 
	by (unfold polyGe_def, auto)
      let ?exp = "\<lambda> s. (polyEval \<beta>  (foldr polySum (map (\<lambda> ct. polyMult (fst ct)  (\<guillemotleft>\<alpha>, snd ct\<guillemotright>)) (zip list (bef  @ s # aft))) (Poly 0 [])))"
      show "?exp s \<ge> ?exp t"
      proof (induct bef arbitrary: list)
	case Nil
	thus ?case using stb by (cases list, auto)
      next
	case (Cons b befs)
	thus ?case by (cases list, auto)
      qed
    qed
  qed
qed

fun getSub where "getSub (Sub \<sigma>) = \<sigma>"

lemma polySub[simp] : "polyEval \<beta> (\<guillemotleft>\<alpha>, (t \<cdot> \<sigma>)\<guillemotright>) = polyEval (\<lambda> x. polyEval \<beta> (\<guillemotleft>\<alpha>, (getSub \<sigma> x)\<guillemotright>)) (\<guillemotleft>\<alpha>, t\<guillemotright>)" 
proof (cases \<sigma>)
  case (Sub sig)
  have "polyEval \<beta> (\<guillemotleft>\<alpha>, t \<cdot> Sub sig\<guillemotright>) = polyEval (\<lambda> x. polyEval \<beta> (\<guillemotleft>\<alpha>, sig x\<guillemotright>)) (\<guillemotleft>\<alpha>, t\<guillemotright>)" (is "_ = polyEval ?B (\<guillemotleft>\<alpha>, t\<guillemotright>)")
  proof (induct t)
    case (Var x)
    thus ?case by simp
  next
    case (Fun f ss)
    let ?ts = "map (\<lambda> s. s \<cdot> Sub sig) ss"
    show ?case
    proof (cases "\<alpha> f")
      case (Pair c list)      
      hence "polyEval \<beta> (\<guillemotleft>\<alpha>, Fun f ss \<cdot> Sub sig\<guillemotright>) = c + 
                polyEval \<beta> ((foldr polySum
               (map (\<lambda> ct. (fst ct) * (\<guillemotleft>\<alpha>, snd ct\<guillemotright>)) (zip list ?ts))                 
      (Poly 0 [])))" by simp
      also have "\<dots> = c + polyEval ?B ((foldr polySum (map (\<lambda> ct. (fst ct) * (\<guillemotleft>\<alpha>, snd ct\<guillemotright>)) (zip list ss)) (Poly 0 [])))" (is "c + ?left = c + ?right")
      proof -
	have "?left = ?right" using Fun
	proof (induct ss arbitrary:list, simp)
	  case (Cons s ss)	  
	  thus ?case by (cases list, auto)
	qed
	thus ?thesis by simp
      qed
      finally show ?thesis using Pair by simp
    qed
  qed
  with Sub show ?thesis by simp
qed

lemma poloSSub[simp] : "sub_closed (poloS \<alpha>)" (is "sub_closed ?P")
proof (unfold sub_closed_def, rule subsetI)
  fix stSig
  assume stSig: "stSig \<in> CS ?P"
  thus "stSig \<in> ?P"
  proof (cases stSig)
    case (Pair sSig tSig)
    with stSig have "(sSig, tSig) \<in> CS ?P" by simp
    hence "(sSig, tSig) \<in> ?P" 
    proof
      fix s t \<sigma>
      assume sSig: "sSig = s \<cdot> \<sigma>" and tSig: "tSig = t \<cdot> \<sigma>" and "(s,t) \<in> ?P"
      hence st: "\<guillemotleft>\<alpha>, s\<guillemotright> > \<guillemotleft>\<alpha>, t\<guillemotright>" by (auto simp: poloS_def poloNS_def)
      have "\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright> > \<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>" 
      proof (unfold polyGt_def, rule allI)
	fix \<beta>
	from st
	show "polyEval \<beta> (\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright>) > polyEval \<beta> (\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>)" by (unfold polyGt_def, simp)
      qed
      with sSig tSig show ?thesis by (simp add: poloS_def poloNS_def)
    qed
    with Pair show ?thesis by simp
  qed
qed

lemma poloNSSub[simp] : "sub_closed (poloNS \<alpha>)" (is "sub_closed ?P")
proof (unfold sub_closed_def, rule subsetI)
  fix stSig
  assume stSig: "stSig \<in> CS ?P"
  thus "stSig \<in> ?P"
  proof (cases stSig)
    case (Pair sSig tSig)
    with stSig have "(sSig, tSig) \<in> CS ?P" by simp
    hence "(sSig, tSig) \<in> ?P" 
    proof
      fix s t \<sigma>
      assume sSig: "sSig = s \<cdot> \<sigma>" and tSig: "tSig = t \<cdot> \<sigma>" and "(s,t) \<in> ?P"
      hence st: "\<guillemotleft>\<alpha>, s\<guillemotright> \<ge> \<guillemotleft>\<alpha>, t\<guillemotright>" by (auto simp: poloS_def poloNS_def)
      have "\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright> \<ge> \<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>" 
      proof (unfold polyGe_def, rule allI)
	fix \<beta>
	from st
	show "polyEval \<beta> (\<guillemotleft>\<alpha>, s \<cdot> \<sigma>\<guillemotright>) \<ge> polyEval \<beta> (\<guillemotleft>\<alpha>, t \<cdot> \<sigma>\<guillemotright>)" by (unfold polyGe_def, simp)
      qed
      with sSig tSig show ?thesis by (auto simp: poloS_def poloNS_def)
    qed
    with Pair show ?thesis by simp
  qed
qed

(*
end locale 
*)



(* theory is done, so lets prove that every polynomial order forms a reduction pair *)

interpretation polyRedpair:  redpair [ "poloS \<alpha>" "poloNS \<alpha>" ]
by (unfold_locales, auto simp: 
   poloSN 
   poloCompat 
   poloCC
   poloSSub
   poloNSSub)

(* now we come to checking that two terms are in relation  *)

(*locale polyorderImpl = polyorder
begin
*)

fun checkVPLists :: "(carrier \<times> 'v) list \<Rightarrow> (carrier \<times> 'v) list \<Rightarrow> 'v monad"
  where "checkVPLists _ [] = OK"
     |  "checkVPLists vp ((n,x) # vq) = (if n = 0 then checkVPLists vp vq else (case lookupVar x vp of None \<Rightarrow> error x | Some (m, vpp) \<Rightarrow> checkB (m \<ge> n) x \<guillemotright> checkVPLists vpp vq))"


fun checkPoly :: "('v :: sshow) poly \<Rightarrow> 'v poly \<Rightarrow> sstring list monad"
  where "checkPoly (Poly c vp) (Poly d vq) = (
   checkB (c \<ge> d)
      [''the constant of the left polynomial is too small'']
  \<guillemotright> checkVPLists vp vq 
     \<leftarrow> (\<lambda> x. [ ''variable '' +++ sshow x +++  '' is problematic'']))"



lemma checkPolySound[simp] : "isOK(checkPoly p q) \<Longrightarrow> p \<ge> q"
proof -
  assume check: "isOK(checkPoly p q)"
  show "p \<ge> q"
  proof (cases p)
    case (Poly c vp)
    show ?thesis
    proof (cases q)
      case (Poly d vq)
      show ?thesis
      proof (cases "c \<ge> d")
	case False with prems show ?thesis by simp
      next
	case True      
	{ assume check2: "isOK(checkVPLists vp vq)"
	  with True 
	  have "Poly c vp \<ge> Poly d vq" 
	  proof (unfold polyGe_def, clarify)
	    fix \<beta>
	    from check2 have "(\<Sum>nv\<leftarrow>vq. fst nv * \<beta> (snd nv))
              \<le> (\<Sum>nv\<leftarrow>vp. fst nv * \<beta> (snd nv))" 
	    proof (induct vq arbitrary: vp)
	      case Nil thus ?case by simp
	    next
	      case (Cons nx vqq)
	      show ?case 
	      proof (cases nx)
		case (Pair n x)
		with Cons show ?thesis 
		proof (cases "n = 0", simp)
		  case False
		  with Cons Pair show ?thesis proof (cases "lookupVar x vp", simp)
		    case (Some mvpp)
		    show ?thesis 
		    proof (cases mvpp)
		      case (Pair m vpp)
		      show ?thesis 
		      proof (cases "m \<ge> n")
			case False with Cons Pair Some `n \<noteq> 0` `nx = (n,x)` show ?thesis by simp
		      next
			case True with Cons Pair Some `n \<noteq> 0` `nx = (n,x)` have "isOK(checkVPLists vpp vqq)" by simp
			with Cons have ind: "(\<Sum>nv\<leftarrow>vqq. fst nv * \<beta> (snd nv)) \<le> (\<Sum>nv\<leftarrow>vpp. fst nv * \<beta> (snd nv))" by simp
			from Some Pair have lookup: "lookupVar x vp = Some (m,vpp)" by simp
			have "(\<Sum>nv\<leftarrow>nx # vqq. fst nv * \<beta> (snd nv)) \<le> n * \<beta> x + (\<Sum>nv\<leftarrow> vqq. fst nv * \<beta> (snd nv))" using `nx = (n,x)` by simp
			also have "\<dots> \<le> n * (\<beta> x) + (\<Sum>nv\<leftarrow>vpp. fst nv * \<beta> (snd nv))" using ind by simp
			also have "\<dots> \<le> m * (\<beta> x) + (\<Sum>nv\<leftarrow>vpp. fst nv * \<beta> (snd nv))" using True by simp
			also have "\<dots> \<le> (\<Sum>nv\<leftarrow>vp. fst nv * \<beta> (snd nv))" (is "?left \<le> ?right") 
			proof -
			  from lookup have "?left = ?right" 
			  proof (induct vp arbitrary: m vpp)
			    case Nil thus ?case by simp
			  next
			    case (Cons cy vp)
			    show ?case
			    proof (cases cy)
			      case (Pair c y)
			      with Cons show ?thesis proof (cases "x = y", simp)
				case False
				with Cons Pair show ?thesis proof (cases "lookupVar x vp", simp)
				  case (Some dvpp)
				  with False Pair Cons show ?thesis by (cases dvpp, auto)
				qed
			      qed
			    qed
			  qed
			  thus ?thesis by simp
			qed
			finally show ?thesis .
		      qed
		    qed
		  qed
		qed
	      qed
	    qed
	    with True show "polyEval \<beta> (Poly d vq) \<le> polyEval \<beta> (Poly c vp)" by simp
	  qed	  
	}
	with prems show ?thesis by simp
      qed
    qed
  qed
qed

declare checkPoly.simps[simp del]

fun genIndices :: "nat \<Rightarrow> int \<Rightarrow> ('f,sstring) term  list"
  where "genIndices 0 _ = []" 
  |  "genIndices (Suc n) i = (Var (''x_'' +++ (sshow i)) # genIndices n (i+1))"
fun genTerm :: "'f \<Rightarrow> nat \<Rightarrow> ('f,string)term" where "genTerm f n = Fun f (genIndices n 0)"
fun "showInterpretation" :: "'f polyinter \<Rightarrow> 'f \<Rightarrow> sstring"
  where "showInterpretation \<alpha> f = (sshow (\<guillemotleft>\<alpha>, genTerm f (length (snd (\<alpha> f)))\<guillemotright>))"

definition checkPolyGe :: "('v :: sshow) poly \<Rightarrow> ('v :: sshow ) poly \<Rightarrow> sstring list monad"
  where "checkPolyGe p q \<equiv> 
          checkPoly p q
  \<leftarrow> (\<lambda> s. (sConcat [''Could not orient '', sshow p, '' >= '', sshow q]) # s) "

definition checkPolyGt :: "('v :: sshow) poly \<Rightarrow> ('v :: sshow) poly \<Rightarrow> sstring list monad"
  where "checkPolyGt p q \<equiv> 
          checkPoly p (polyOne + q)
  \<leftarrow> (\<lambda> s. (sConcat [''Could not orient '', sshow p, '' > '', sshow q]) # s) "


lemma checkPolyGeSound[simp] : "isOK(checkPolyGe p q) \<Longrightarrow> p \<ge> q" using assms unfolding checkPolyGe_def by simp
lemma checkPolyGtSound[simp] : "isOK(checkPolyGt p q) \<Longrightarrow> p > q"
using assms
proof -
  assume "isOK(checkPolyGt p q)" 
  hence "p \<ge> polyOne + q" unfolding checkPolyGt_def by simp
  hence ge: "\<forall> \<beta>. polyEval \<beta> p \<ge> 1 + polyEval \<beta> q" unfolding polyGe_def by simp
  have "\<forall> \<beta>. polyEval \<beta> p > polyEval \<beta> q" 
  proof
    fix \<beta> 
    from ge have "polyEval \<beta> p \<ge> 1 + polyEval \<beta> q" ..
    thus "polyEval \<beta> p > polyEval \<beta> q" by simp
  qed
  thus ?thesis by (auto simp: polyGt_def )
qed




definition checkNSPolo :: "'f polyinter \<Rightarrow> ('f :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad"
  where "checkNSPolo \<alpha> lr \<equiv> checkPolyGe (\<guillemotleft>\<alpha>, fst lr\<guillemotright>) (\<guillemotleft>\<alpha>, snd lr\<guillemotright>) 
      \<leftarrow> (\<lambda> s. (sConcat [ ''Could not orient '' , sshow (fst lr), '' >= '', sshow (snd lr)]) # s)
"

lemma checkNSPoloSound[simp] : "isOK(checkNSPolo \<alpha> lr) \<Longrightarrow> lr \<in> poloNS \<alpha>"
  unfolding checkNSPolo_def by (cases lr, auto simp: poloS_def poloNS_def)


definition checkSPolo :: "'f polyinter \<Rightarrow> ('f :: sshow ,'v :: sshow)rule \<Rightarrow> sstring list monad"
  where "checkSPolo \<alpha> lr \<equiv> checkPolyGt (\<guillemotleft>\<alpha>, fst lr\<guillemotright>) (\<guillemotleft>\<alpha>, snd lr\<guillemotright>)
     \<leftarrow> (\<lambda> s. (sConcat [ ''Could not orient '' , sshow (fst lr), '' > '', sshow (snd lr)]) # s)
"

lemma checkSPoloSound[simp] : "isOK(checkSPolo \<alpha> lr) \<Longrightarrow> lr \<in> poloS \<alpha>"
  unfolding checkSPolo_def by (cases lr, auto simp: poloS_def poloS_def)

fun createPolyInter :: "('f \<times> (nat \<times> nat list)) list \<Rightarrow> 'f polyinter"
  where "createPolyInter [] f = (0,[])"
     |  "createPolyInter ((g,i) # rest) f = (if f = g then i else createPolyInter rest f)"


fun createRepresentation :: "('f :: sshow) polyinter \<Rightarrow> 'f list \<Rightarrow> sstring list"
  where "createRepresentation \<alpha> fs =  (''polynomial interpretation: '' #
             map (\<lambda> g . sConcat [''Pol('', sshow g,  '') = '', showInterpretation \<alpha> g]) (remdups fs))"
(*
end 
*)

datatype 'f RedPair = Polo "('f \<times> (nat \<times> nat list)) list"

fun createRedPairProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
        sstring list monad"
where "createRedPairProc (Polo interpr) P rP R = (let \<alpha> = createPolyInter interpr in 
  checkRedPairProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>) 
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map fst interpr)))
           P rP R)" 



lemma createRedPairProcSound[simp] : fixes P rP R
assumes check: "isOK(createRedPairProc redp P rP R)"
   shows "finite_dpp (set rP, set R) \<Longrightarrow> finite_dpp (set P, set R)"
using assms 
proof (cases redp, simp add: Let_def) 
  case (Polo interpr)
  assume fin: "finite_dpp (set rP, set R)"
  let ?succ = "poloS (createPolyInter interpr)"
  let ?succsim = "poloNS (createPolyInter interpr)"
  let ?checkS = "checkSPolo (createPolyInter interpr)"
  let ?checkNS = "checkNSPolo (createPolyInter interpr)"
  show ?thesis 
  proof (cases "computeSplit (\<lambda> dp. isOK(?checkS dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check Polo have cond:
      "set P - set NSp \<subseteq> ?succ \<and> 
      ?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ   \<and> 
      set R \<subseteq> ?succsim     \<and>
      set NSp \<subseteq> set rP"
      by (auto simp: Let_def mem_iff)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> ?succ" by (auto simp: mem_iff)
    from Polo fin have fin2: "finite_dpp (?NSp, set R)" unfolding finite_dpp_def by (auto simp:mem_iff)
    show "finite_dpp (set P, set R)" 
    proof (rule polyRedpair.redPairProcessorDpp)
      show "finite_dpp (?NSp, set R)" using fin2 .
    next
      from stri show "set P - ?NSp \<subseteq> ?succ" .
    next
      from cond id show "?NSp \<inter> set P \<subseteq> ?succsim \<union> ?succ" by simp
    next
      from cond show "set R \<subseteq> ?succsim" by simp
    qed
  qed
qed
    
declare createRedPairProc.simps[simp del]


end




