(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Util imports Main begin

lemma Union_map_iff_UN_lt_length: "(\<Union>(set (map f xs))) = (\<Union>i < length xs. f (xs!i))"
proof (induct xs)
 case Nil thus ?case by auto
next
 case (Cons x xs) show ?case
 proof (intro equalityI subsetI)
  fix a assume "a \<in> \<Union>set (map f (x#xs))"
  hence "a \<in> f x \<or> a \<in> \<Union>set (map f xs)" by simp
  thus "a \<in> (\<Union>i < length (x#xs). f ((x#xs)!i))"
  proof
   assume "a \<in> f x" thus ?thesis by force
  next
   assume "a \<in> \<Union>set (map f xs)"
   with Cons have "a \<in> (\<Union>i < length xs. f (xs!i))" by simp
   then obtain i where "i < length xs" and "a \<in> f (xs!i)" unfolding UN_iff by best
   thus ?thesis by force
  qed
 next
  fix a assume "a \<in> (\<Union>i < length (x#xs). f ((x#xs)!i))"
  then obtain i where "i < length (x#xs)" and "a \<in> f ((x#xs)!i)" by best
  show "a \<in> \<Union>set (map f (x#xs))"
  proof (cases i)
   assume "i = 0" with `a \<in> f ((x#xs)!i)` have "a \<in> f x" by simp
   thus ?thesis by simp
  next
   fix j assume "i = Suc j" 
   with `i < length (x#xs)` and `a \<in> f ((x#xs)!i)`
   have "j < length xs" and "a \<in> f (xs!j)" by auto
   hence "\<exists>i < length xs. a \<in> f (xs!i)" by auto
   hence "a \<in> (\<Union>i < length xs. f (xs!i))" by auto
   with Cons have "a \<in> \<Union>set (map f xs)" by blast
   thus ?thesis by auto
  qed
 qed
qed

lemma take_is_nth: "i \<le> length xs \<Longrightarrow> ((take i xs)@(y#ys))!i = y"
proof (induct xs arbitrary: i)
 case Nil thus ?case by simp
next
 case (Cons x xs)
 thus ?case
 proof (induct i)
  case 0 thus ?case by auto
 next
  case (Suc i)
  from Cons have IH: "i \<le> length xs \<Longrightarrow> ((take i xs)@(y#ys))!i = y" by auto
  from `Suc i \<le> length (x#xs)` have "i \<le> length xs" by simp
  with IH have "((take i xs)@(y#ys))!i = y" by simp
  thus ?case by auto
 qed
qed

lemma nth_take_is_nth: "\<lbrakk>i < j;j \<le> length xs\<rbrakk> \<Longrightarrow> (take j xs @ ys)!i = xs!i"
proof (induct xs arbitrary: i j)
 case Nil thus ?case by auto
next
 case (Cons x xs) thus ?case
 proof (induct j)
  case 0 thus ?case by auto
 next
  case (Suc j) thus ?case
  proof (induct i)
   case 0 thus ?case by auto
  next
   case (Suc i) thus ?case by auto
  qed
 qed
qed

lemma gt_nth_take_drop_is_nth: "\<lbrakk>j < i; j \<le> length xs; i \<le> length xs\<rbrakk> \<Longrightarrow> (take j xs @ y # drop (Suc j) xs)!i = xs!i"
proof (induct xs arbitrary: i j)
 case Nil thus ?case by auto
next
 case (Cons x xs) thus ?case
 proof (induct i)
  case 0 thus ?case by auto
 next
  case (Suc i) thus ?case
  proof (induct j)
   case 0 thus ?case by auto
  next
   case (Suc j) thus ?case by auto
  qed
 qed
qed

lemma nth_take_drop_is_nth: 
 assumes "i \<le> length xs" and "j \<le> length xs" and "i \<noteq> j" 
 shows "(take j xs @ y # drop (Suc j) xs)!i = xs!i"
proof -
 from assms have "i < j \<or> i > j" by auto
 thus ?thesis
 proof
  assume "i < j" thus ?thesis using assms by (auto simp: nth_take_is_nth)
 next
  assume "j < i" thus ?thesis using assms by (auto simp: gt_nth_take_drop_is_nth)
 qed
qed
  
lemma take_drop_imp_nth: "\<lbrakk>take i ss @ x # drop (Suc i) ss = ss\<rbrakk> \<Longrightarrow> x = ss!i"
proof (induct ss arbitrary: i)
 case Nil thus ?case by auto
next
 case (Cons s ss)
 from `take i (s#ss) @ x # drop (Suc i) (s#ss) = (s#ss)` show ?case
 proof (induct i)
  case 0 thus ?case by auto
 next
  case (Suc i)
  from Cons have IH: "take i ss @ x # drop (Suc i) ss = ss \<Longrightarrow> x = ss!i" by auto
  from Suc have "take i ss @ x # drop (Suc i) ss = ss" by auto
  with IH show ?case by auto
 qed
qed
  
lemma map_map_distr:
 assumes "!!x. x \<in> set xs \<Longrightarrow> f (g x) (map g ys) = g (f x ys)"
 shows "map (\<lambda>x. f x (map g ys)) (map g xs) = map g (map (\<lambda>x. f x ys) xs)"
using assms proof (induct xs)
 case Nil thus ?case by simp
next
 case (Cons x xs)
 have "map (\<lambda>x. f x (map g ys)) (map g (x#xs)) = map (\<lambda>x. f x (map g ys)) (g x # map g xs)" by simp
 also have "\<dots> = f (g x) (map g ys) # map (\<lambda>x. f x (map g ys)) (map g xs)" by simp
 also have "\<dots> = g (f x ys) # map (\<lambda>x. f x (map g ys)) (map g xs)" using Cons by simp
 also have "\<dots> = g (f x ys) # map g (map (\<lambda>x. f x ys) xs)" using Cons by simp
 finally show ?case by simp
qed

lemma map_map_distr_assoc:
 fixes ys :: "('a \<times> 'b) list"
 assumes "!!x. x \<in> set xs \<Longrightarrow> f (g x) (map (\<lambda>(k, v). (k, g v)) ys) = g (f x ys)"
 shows "map (\<lambda>x. f x (map (\<lambda>(k,v). (k, g v)) ys)) (map g xs) = map g (map (\<lambda>x. f x ys) xs)"
using assms proof (induct xs)
 case Nil thus ?case by simp
next
 case (Cons x xs)
 let ?g = "\<lambda>(k,v). (k, g v)"
 have "map (\<lambda>x. f x (map ?g ys)) (map g (x#xs)) = map (\<lambda>x. f x (map ?g ys)) (g x # map g xs)" by simp
 also have "\<dots> = f (g x) (map ?g ys) # map (\<lambda>x. f x (map ?g ys)) (map g xs)" by simp
 also have "\<dots> = g (f x ys) # map (\<lambda>x. f x (map ?g ys)) (map g xs)" using Cons by simp
 also have "\<dots> = g (f x ys) # map g (map (\<lambda>x. f x ys) xs)" using Cons by simp
 finally show ?case by simp
qed

lemma bound_nat_induct[consumes 1]:
 "\<lbrakk> n \<in> {l..u}; P l; !!n. \<lbrakk>P n; n \<in> {l..<u}\<rbrakk> \<Longrightarrow> P (Suc n)\<rbrakk> \<Longrightarrow> P n"
proof (induct n rule: nat.induct)
 case zero thus ?case by (induct l) auto
next
 case (Suc n) thus ?case
 proof (cases "Suc n = l")
  case True with `P l` show ?thesis by auto
 next
  case False with `Suc n \<in> {l..u}` have "n \<in> {l..<u}" by auto
  with Suc have "P n" by auto
  from `P n` and `n \<in> {l..<u}` and `\<lbrakk>P n; n \<in> {l..<u}\<rbrakk> \<Longrightarrow> P (Suc n)`
  show ?thesis by auto
 qed
qed

lemma bound_nat_induct':
 "\<lbrakk>P l; !!n. \<lbrakk>P n; n \<in> {l..<u}\<rbrakk> \<Longrightarrow> P (Suc n)\<rbrakk> \<Longrightarrow> \<forall>n \<in> {l..u}. P n"
proof
 fix n
 assume "n \<in> {l..u}" "P l" "!!n. \<lbrakk>P n; n \<in> {l..<u}\<rbrakk> \<Longrightarrow> P (Suc n)" 
 thus "P n" by (rule bound_nat_induct)
qed


fun testPos :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> bool"
where "testPos _ [] = False"
    | "testPos f (x # xs) = (f x \<or> testPos f xs)"

(* Works
export_code testPos in Haskell file "Checker"
*)

lemma testPosSound : "testPos f xs = (\<exists> x. x mem xs \<and> f x)"
  by (induct xs, auto)

lemma testPosSound2 : "testPos f xs = (\<exists> i < length xs. f (xs ! i))"
proof (induct xs, simp)
  case (Cons x xs)
  show ?case 
  proof (cases "f x")
    case True with Cons show ?thesis by auto
  next
    case False 
    show ?thesis
    proof
      assume "testPos f (x # xs)" 
      with False Cons show "\<exists> i < length (x # xs). f ((x # xs) ! i)" by auto
    next
      assume "\<exists> i < length (x # xs). f ((x # xs) ! i)" 
      from this obtain i where "i < length (x # xs)" and "f ((x # xs) ! i)" by auto
      thus "testPos f (x # xs)" using False Cons by (cases i, auto)
    qed
  qed
qed  

lemma testPosCong[fundef_cong] : fixes xs ys f g 
  assumes xs: "xs = ys" and set: "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x" 
  shows "testPos f xs = testPos g ys" 
proof -
  from xs have "testPos f xs = testPos f ys" by simp
  also have "\<dots> = (\<exists> x. x mem ys \<and> f x)" by (rule testPosSound)
  also have "\<dots> = (\<exists> x. x mem ys \<and> g x)" using mem_iff set by best
  also have "\<dots> = testPos g ys" using testPosSound by auto
  finally show ?thesis .
qed


fun testAll :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> bool"
where "testAll f xs = (\<not> testPos (\<lambda> x. \<not> f x) xs)"

(* Works
export_code testAll in Haskell file "Checker"
*)


lemma testAllSound[simp] : "testAll f xs = (\<forall> x \<in> set xs. f x)"
proof -
  from testPosSound have "testAll f xs = (\<not> (\<exists> x. x mem xs \<and> (\<lambda> x. \<not> f x) x))" by auto
  thus ?thesis by (auto simp: mem_iff)
qed

declare testAll.simps[simp del]


(* testing functions with 'a option as return type *)

fun testPosW :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a option"
where "testPosW _ [] = None"
    | "testPosW f (x # xs) = (if f x then Some x else testPosW f xs)"


lemma testPosWSoundNone1: "(testPosW f xs = (None :: 'a option)) \<Longrightarrow> (\<not>(\<exists> (x :: 'a). x mem xs \<and> f x))"
proof (induct xs, simp)
 case (Cons x xs)
  assume none:"testPosW f (x # xs) = None"
  hence "\<not> (f x)" by auto
  with Cons show ?case by auto
qed

lemma testPosWSoundNone2: "(\<not>(\<exists> x. x mem xs \<and> f x)) \<Longrightarrow> (testPosW f xs = None)"
proof (induct xs)
 case Nil
 show ?case by simp
next
 case (Cons x xs)
 assume "\<not>(\<exists> y. y mem (x # xs) \<and> f y)" hence "\<not> (f x)" by auto 
 with Cons show ?case by auto
qed

lemma testPosWSoundNone: "(\<not>(\<exists> x. x mem xs \<and> f x)) = (testPosW f xs = None)"
 using testPosWSoundNone1 testPosWSoundNone2 by auto


lemma testPosWSoundSome1: 
 fixes f xs x
 shows "(testPosW f xs = Some x) \<Longrightarrow> (\<exists> i < length xs. f (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i. \<not> f (xs ! j)))"
proof(induct xs, simp)
 case (Cons y ys)
  show ?case 
  proof (cases "f y")
   case True 
    with Cons show ?thesis by auto
  next
   case False 
    show ?thesis
    proof-
     from `testPosW f (y # ys) = Some x` False Cons 
     have "\<exists> i < length ys. f (ys ! i) \<and> x = ys ! i \<and> (\<forall> j < i. \<not> f (ys ! j))" by auto
      then obtain i where "i < length ys" and "f (ys ! i)" and " x = ys ! i" and min:"(\<forall> j < i. \<not> f (ys ! j))" by auto
      hence "Suc i < length  (y # ys)" and "f ((y # ys) ! (Suc i))" and " x = (y # ys) ! (Suc i)" by auto
      have "\<forall> j < (Suc i). \<not> f ((y # ys) ! j)"
       proof 
        fix j
        show "j < Suc i \<longrightarrow> \<not> f ((y # ys) ! j)"
        proof (cases j)
         case 0 
          from False and 0 show "j < Suc i \<longrightarrow> \<not> f ((y # ys) ! j)" by auto
        next
         case (Suc k)
          show "j < Suc i \<longrightarrow> \<not> f ((y # ys) ! j)"
          proof
            assume "j < Suc i"
            with Suc have "k < i" by simp
            with min have "\<not> f (ys ! k)" by auto
            hence " \<not> f ((y # ys) ! (Suc k))" by simp
            with Suc show "\<not> f ((y # ys) ! j)" by auto
          qed
        qed
       qed
      with `Suc i < length  (y # ys)` and `f ((y # ys) ! (Suc i))` and `x = (y # ys) ! (Suc i)` 
      show "\<exists>i<length (y # ys). f ((y # ys) ! i) \<and> x = (y # ys) ! i \<and> (\<forall>j<i. \<not> f ((y # ys) ! j))" by auto
    qed
  qed
qed
      
    

lemma testPosWSoundSome2: 
 "(\<exists> i < length xs. f (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i. \<not> f (xs ! j))) \<Longrightarrow> (testPosW f xs = Some x)" 
proof (induct xs, simp)
 case (Cons y ys)
  with `\<exists> i < length (y # ys). f ((y # ys) ! i) \<and> (x =  ((y # ys) ! i)) \<and> (\<forall> j < i. \<not> f ((y # ys) ! j))`
  obtain i where "i < length (y # ys)" and "f ((y # ys) ! i)" 
           and "x = ((y # ys) ! i)" and "\<forall> j < i. \<not> f ((y # ys) ! j)" by auto
  show ?case
  proof (cases i, simp)
   case 0
    with `f ((y # ys) ! i)` have "f y" by simp
    with  `x = (y # ys) ! i` 0 have "y = x" by simp
    with `f y` show " (f y \<longrightarrow> y = x) \<and> (\<not> f y \<longrightarrow> testPosW f ys = Some x)" by auto
  next
   case (Suc k)
    from Suc `\<forall> j < i. \<not> f ((y # ys) ! j)` have "\<not> f y" by auto
    hence rec:"testPosW f (y # ys) = testPosW f ys" by auto
    from Suc `i < length (y # ys)` have k1:"k < length ys" by simp
    from Suc `f ((y # ys) ! i)` have k2:"f (ys ! k)" by simp
    from Suc `x = (y # ys) ! i` have k3:"x = (ys ! k)" by simp
    from Suc `\<forall> j < i. \<not> f ((y # ys) ! j)` have "\<forall> j < k. \<not> f (ys ! j)" by auto
    with k1 k2 k3 have "\<exists>i<length ys. f (ys ! i) \<and> x = ys ! i \<and> (\<forall>j<i. \<not> f (ys ! j))" by auto
    with Cons have "testPosW f ys = Some x" by auto
    with rec show "testPosW f (y # ys) = Some x" by simp
  qed
qed



lemma testPosWSoundSome:
 "(\<exists> i < length xs. f (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i. \<not> f (xs ! j))) = (testPosW f xs = Some x)"
proof
 assume "(\<exists> i < length xs. f (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i. \<not> f (xs ! j)))"
 from this show "testPosW f xs = Some x" by (rule testPosWSoundSome2[where ?xs=xs])
next
 assume "testPosW f xs = Some x"
 thus "(\<exists> i < length xs. f (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i. \<not> f (xs ! j)))" by (rule testPosWSoundSome1[where ?xs=xs]) 
qed


lemma testPosWCong[fundef_cong] : fixes xs ys f g 
  assumes xs: "xs = ys" and set: "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x" 
  shows "testPosW f xs = testPosW g ys"
proof (cases "testPosW f xs")
  case None
   hence "testPosW f ys = None" using xs by simp
   hence "\<not>(\<exists> x. x mem ys \<and> f x)" using testPosWSoundNone1 by auto
   hence "\<not>(\<exists> x. x mem ys \<and> g x)" using mem_iff set by best   
   hence "testPosW g ys = None"  using testPosWSoundNone2 by auto
   thus "testPosW f xs = testPosW g ys" by auto
 next
  case (Some x) 
   hence "testPosW f ys = Some x" using xs by simp
   hence "\<exists> i < length ys. f (ys ! i) \<and> (x =  (ys ! i)) \<and> (\<forall> j < i. \<not> f (ys ! j))" by (rule testPosWSoundSome1[where ?xs=ys])
   hence "\<exists> i < length ys. g (ys ! i) \<and> (x =  (ys ! i)) \<and> (\<forall> j < i. \<not> g (ys ! j))" using mem_iff set by auto
   hence "testPosW g ys = Some x" by (rule testPosWSoundSome2) 
   thus "testPosW f xs = testPosW g ys" by auto
qed 



fun testAllW :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a option"
 where "testAllW f xs = (testPosW (\<lambda> x. \<not> f x) xs)" 

lemma testAllWSoundNone[simp] : "(testAllW f xs = None) = (\<forall> x \<in> set xs. f x)"
proof
 assume "testAllW f xs = None"
 hence "testPosW (\<lambda> x. \<not> f x) xs = None" by simp
 hence "\<not> (\<exists> x. x mem xs \<and> (\<not> f x))" by (rule testPosWSoundNone1)
 thus "\<forall> x \<in> set xs. f x" using mem_iff by best
next
 assume "\<forall> x \<in> set xs. f x"
 hence "(\<not> (\<exists> x. x mem xs \<and> (\<not> f x)))" using mem_iff by best  
 hence "testPosW (\<lambda> x. \<not> f x) xs = None" by (rule testPosWSoundNone2)
 thus "testAllW f xs = None" by auto 
qed

lemma testAllWSoundSome[simp] : "(testAllW f xs = Some x) = (\<exists> i < length xs.  \<not> f (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i. f (xs ! j)))"
proof
 assume "testAllW f xs = Some x"
 hence "testPosW (\<lambda> x. \<not> f x) xs = Some x" by simp 
 hence "\<exists> i < length xs. (\<lambda> x. \<not> f x) (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i.  \<not>((\<lambda> x. \<not> f x) (xs ! j)))" by (rule testPosWSoundSome1) 
 thus "\<exists> i < length xs.  \<not> f (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i. f (xs ! j))" by auto
next
 assume "\<exists> i < length xs.  \<not> f (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i. f (xs ! j))"
 hence  "\<exists> i < length xs. (\<lambda> x. \<not> f x) (xs ! i) \<and> (x =  (xs ! i)) \<and> (\<forall> j < i.  \<not>((\<lambda> x. \<not> f x) (xs ! j)))" by auto
 hence  "testPosW (\<lambda> x. \<not> f x) xs = Some x" by (rule testPosWSoundSome2)
 thus  "testAllW f xs = Some x" by simp
qed


declare testAllW.simps[simp del]

fun testAllW2 :: "('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a option"
where "testAllW2 f g xs rem = (testPosW  (\<lambda> x. if x mem rem then \<not> g x else \<not> f x) xs)"

lemma testAllW2SoundNone : "(testAllW2 f g alll rem = None) =((\<forall> x \<in> set alll - set rem. f x) \<and> (\<forall> x \<in> set alll \<inter> set rem. g x))"
proof
 assume none:"testAllW2 f g alll rem = None"
let ?p = "(\<lambda> x. if x mem rem then \<not> g x else \<not> f x)"
 from none have "testPosW (\<lambda> x. if x mem rem then \<not> g x else \<not> f x) alll = None" by simp
 hence nex:"(\<not> (\<exists> x. x mem alll \<and> (?p x)))" using testPosWSoundNone1[where ?f = ?p and ?xs = alll] by simp
 have all:"\<forall> x \<in> set alll - set rem. f x"
 proof
  fix x
  assume xin:"x \<in> (set alll - set rem)"
  thus "f x"
  proof
   from xin have "x \<in> set alll" and "x \<notin> set rem" by auto 
   with nex have "\<not> ?p x" using mem_iff by best
   hence "if x mem rem then g x else f x" by auto
   from `x \<notin> set rem` have "\<not> (x mem rem)" using mem_iff by best
   with `if x mem rem then g x else f x` show "f x" by auto
  qed
 qed
 have "\<forall> x \<in> set alll \<inter> set rem. g x"
 proof
  fix x
  assume xin:"x \<in> (set alll \<inter> set rem)"
  thus "g x"
  proof
   from xin have "x \<in> set alll" and "x \<in> set rem" by auto
   with nex have "\<not> ?p x" using mem_iff by best
   hence "if x mem rem then g x else f x" by auto
   from `x \<in> set rem` have "x mem rem" using mem_iff by best
   with `if x mem rem then g x else f x` show "g x" by auto
  qed
 qed
 with all show "(\<forall> x \<in> set alll - set rem. f x) \<and> (\<forall> x \<in> set alll \<inter> set rem. g x)" by auto
next
 let ?p = "(\<lambda> x. if x mem rem then \<not> g x else \<not> f x)"
 assume all:"(\<forall>x\<in>set alll - set rem. f x) \<and> (\<forall>x\<in>set alll \<inter> set rem. g x)"
 hence "\<forall> x \<in> set alll. \<not> (?p x)"
 proof
  show "\<forall>x\<in>set alll. \<not> (if x mem rem then \<not> g x else \<not> f x)"
  proof
   fix x
   assume "x \<in> set alll"
   show "\<not> (if x mem rem then \<not> g x else \<not> f x)"
   proof (cases "x mem rem")
    case True
     from True `x \<in> set alll` have "x \<in> set alll \<inter> set rem" using mem_iff by best
     with all have "g x" by auto
     with True show "\<not> (if x mem rem then \<not> g x else \<not> f x)"  by auto
   next
    case False
     from False `x \<in> set alll` have "x \<in> set alll - set rem" using mem_iff by best 
     with all have "f x" by auto
     with False show "\<not> (if x mem rem then \<not> g x else \<not> f x)"  by auto     
   qed
  qed
 qed
 hence "\<not> (\<exists> x. x mem alll \<and> (?p x))" using mem_iff by best
 hence "testPosW ?p alll = None" using testPosWSoundNone2[where ?f = ?p and ?xs = alll] by simp
 thus "testAllW2 f g alll rem = None" by auto
qed


lemma testAllW2SoundSome: "(testAllW2 f g alll rem = Some x) = 
 (\<exists> i < length alll. (if (alll ! i) mem rem then \<not> g (alll ! i) else \<not> f (alll ! i)) \<and> (x = alll ! i) \<and> (\<forall> j < i. \<not> (if (alll ! j) mem rem then \<not> g (alll ! j) else \<not> f (alll ! j))))"
proof
 let ?p = "\<lambda> x. if x mem rem then \<not> g x else \<not> f x"
 assume "testAllW2 f g alll rem = Some x"
 hence "testPosW ?p alll = Some x" by simp
 thus "\<exists> i < length alll. ?p (alll ! i) \<and> (x = alll ! i) \<and> (\<forall> j < i.  \<not>(?p (alll ! j)))" by (rule testPosWSoundSome1) 
next
 let ?p = "(\<lambda> x. if x mem rem then \<not> g x else \<not> f x)"
 assume "\<exists> i < length alll. ?p (alll ! i) \<and> (x = alll ! i) \<and> (\<forall> j < i.  \<not>(?p (alll ! j)))"
 hence  "(testPosW ?p alll = Some x)" by (rule testPosWSoundSome2)
 thus  "testAllW2 f g alll rem = Some x" by simp
qed

datatype ('a,'b)either = Left 'a | Right 'b
definition isLeft where "isLeft lr \<equiv> (case lr of Left l \<Rightarrow> True | Right r \<Rightarrow> False)"
fun left where "left (Left l) = l"
fun right where "right (Right r) = r"


fun computeSplit :: "('a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> ('a list \<times> 'a list)"
where "computeSplit _ [] = ([], [])"
   |  "computeSplit f (x # xs) = (let (l,r) = computeSplit f xs in if f x then (x # l, r) else (l, x # r))" 

lemma computeSplitSound[simp] : "computeSplit f xs = ([x \<leftarrow> xs. f x], [x \<leftarrow> xs. \<not> f x])"
proof (induct xs, simp)
  case (Cons x xs) thus ?case by (cases "f x", auto)
qed
declare computeSplit.simps[simp del]

fun unionI :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list"
where "unionI [] ys = ys"
   |  "unionI (x # xs) ys = (let zs = unionI xs ys in if x mem zs then zs else x # zs)"

lemma unionISound[simp] : "set (unionI xs ys) = set xs \<union> set ys"
proof (induct xs, simp)
  case (Cons x xs)
  thus ?case by (cases "x mem (unionI xs ys)", auto simp: mem_iff Let_def)
qed
declare unionI.simps[simp del]


abbreviation mapEither :: "('a \<Rightarrow> 'c) \<Rightarrow> ('b \<Rightarrow> 'c) \<Rightarrow> ('a,'b) either list \<Rightarrow> 'c list"
  where "mapEither f g \<equiv> map (\<lambda> e. case e of Left l \<Rightarrow> f l | Right r \<Rightarrow> g r)"

abbreviation map_filter :: "('a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'a list \<Rightarrow> 'b list"
where "map_filter f g xs \<equiv> foldr (\<lambda> x rest. if f x then g x # rest else rest) xs []"

end
